/*
    Copyright (c) 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <errno.h>

#include "util.h"
#include "timer.h"
#include "log.h"
#include "exceptions.h"

namespace dSA
{

TimerSubscriber::TimerSubscriber(unsigned int notifyInterval, bool once) :
                                 m_notifyInterval(notifyInterval), m_once(once)
{
    timeutil::getTimespecAfterMillis(m_notifyInterval * 1000, &m_nextNotify);
    m_valid = true;
}

TimerSubscriber::TimerSubscriber() : m_notifyInterval(0), m_once(true)
{
    timeutil::getTimespecAfterMillis(m_notifyInterval * 1000, &m_nextNotify);
    m_valid = true;
}

struct timespec *TimerSubscriber::getNextNotify()
{
    return &m_nextNotify;
}

unsigned int TimerSubscriber::getNotifyInterval()
{
    return m_notifyInterval;
}

void TimerSubscriber::notified()
{
    if (m_once)
    {
        m_valid = false;
    }
    timeutil::getTimespecAfterMillis(m_notifyInterval * 1000, &m_nextNotify);
}

bool TimerSubscriber::once()
{
    return m_once;
}

bool TimerSubscriber::valid()
{
    return m_valid;
}

TimerSubscriber::~TimerSubscriber() {}

Timer *Timer::m_instance = NULL;
Timer *Timer::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new Timer();
        if (m_instance == NULL)
        {
            throw _Exception("could not create Timer class");
        }
    }

    return m_instance;
}

Timer *Timer::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("Timer class not initialized");
    }

    return m_instance;
}

Timer::Timer()
{
    int ret = pthread_mutex_init(&m_subscribers_mutex, NULL);
    if (ret != 0)
    {
        throw _Exception("failed to initialize timer subscriber mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_notify_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_subscribers_mutex);
        throw _Exception("failed to initialize notify condition: " +
                         std::string(strerror(errno)));
    }
}

void Timer::addSubscriber(boost::shared_ptr<TimerSubscriber> ts)
{
    if (ts == NULL)
    {
        throw _Exception("tried to add an invalid timer subscriber (NULL)");
    }

    if (ts->getNotifyInterval() <= 0)
    {
        throw _Exception("tried to add timer with illegalinterval");
    }

    FILE_LOG(logDEBUG2) << "adding timer subscriber with interval: "
                        << ts->getNotifyInterval() << " secs, oneshot: "
                        << ts->once();
    lock();

    std::list<boost::shared_ptr<TimerSubscriber> >::iterator i =
                                                        m_subscribers.begin();
    while (i != m_subscribers.end())
    {
        if ((*i) == ts)
        {
            unlock();
            throw _Exception("tried to add same timer twice");
        }

        i++;
    }

    m_subscribers.push_back(ts);
    pthread_cond_signal(&m_notify_cond);
    unlock();
}

void Timer::signal()
{
    pthread_cond_signal(&m_notify_cond);
};

void Timer::wait()
{
    lock();
    if (m_subscribers.size() <= 0)
    {
        FILE_LOG(logDEBUG3) << " timer wait: nothing to do, waiting for "
                            << "subscriber";
        pthread_cond_wait(&m_notify_cond, &m_subscribers_mutex);
    }
    else
    {
        struct timespec *timeout = NULL;
        std::list<boost::shared_ptr<TimerSubscriber> >::iterator i =
                                                        m_subscribers.begin();
        while (i != m_subscribers.end())
        {
            struct timespec *nextNotify = (*i)->getNextNotify();
            if ((timeout == NULL) ||
                (timeutil::compareTimespecs(nextNotify, timeout) > 0))
            {
                timeout = nextNotify;
            }
            i++;
        }

        struct timespec now;
        timeutil::getTimespecNow(&now);
        if (timeutil::compareTimespecs(timeout, &now) < 0)
        {
            int ret = pthread_cond_timedwait(&m_notify_cond,
                                             &m_subscribers_mutex, timeout);
            if (ret != 0 && ret != ETIMEDOUT)
            {
                unlock();
                throw _Exception("timed wait returned error" +
                                 std::string(strerror(errno)));
            }

            if (ret == ETIMEDOUT)
            {
                unlock();
                notifySubscribers();
                lock();
            }
        }
        else
        {
            unlock();
            notifySubscribers();
            lock();
        }
    }
    unlock();
}

void Timer::notifySubscribers()
{
    std::list<boost::shared_ptr<TimerSubscriber> > subscribers;

    lock();
    subscribers.swap(m_subscribers);
    unlock();

    struct timespec now;
    timeutil::getTimespecNow(&now);
    std::list<boost::shared_ptr<TimerSubscriber> >::iterator i =
                                                           subscribers.begin();

    while (i != subscribers.end())
    {
        if (timeutil::compareTimespecs((*i)->getNextNotify(), &now) >= 0)
        {
            FILE_LOG(logDEBUG3) <<  "notifying subscribers";
            try
            {
                (*i)->notify();
                (*i)->notified();
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            catch (...)
            {
                FILE_LOG(logERROR) << "timer notify caught exception";
            }
        }
        if ((*i)->valid())
        {
            lock();
            m_subscribers.push_back((*i));
            unlock();
        }
        i++;
    }
}

void Timer::removeSubscriber(boost::shared_ptr<TimerSubscriber> ts)
{
    lock();
    std::list<boost::shared_ptr<TimerSubscriber> >::iterator i =
                                                        m_subscribers.begin();

    while (i != m_subscribers.end())
    {
        if ((*i) == ts)
        {
            FILE_LOG(logDEBUG2) << "removing timer subscriber";
            i = m_subscribers.erase(i);
            break;
        }

        i++;
    }

    unlock();
}

void Timer::lock()
{
    int ret = pthread_mutex_lock(&m_subscribers_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock timer subscriber mutex: " +
                         std::string(strerror(errno)));
    }
}

void Timer::unlock()
{
    int ret = pthread_mutex_unlock(&m_subscribers_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock timer subscriber mutex: " +
                         std::string(strerror(errno)));
    }
}

Timer::~Timer()
{
    lock();
    while (!m_subscribers.empty())
    {
        m_subscribers.pop_front();
    }
    unlock();

    pthread_mutex_destroy(&m_subscribers_mutex);
    pthread_cond_destroy(&m_notify_cond);
}

}; // namespace
